/* dema.h
 * monitoring tables
 *
 * $Id:  $
 *
 * Xplico System
 * By Gianluca Costa <g.costa@xplico.org>
 * Copyright 2007-2011 Gianluca Costa & Andrea de Franceschi. Web: www.xplico.org
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#ifndef __DEMA_H__
#define __DEMA_H__

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>

#define DM_FILENAME_PATH         2048
#define DM_FILTER_LINE           2048
#define DM_TBL_ADD               500
#define DM_HASH_STR              1024
#define DM_POL_NAME              "pol_"
#define DM_SESSION_NAME          "sol_"
#define DM_POL_DIR               "%s/pol_%i"
#define DM_TMP_DIR               "%s/pol_%i/tmp"
#define DM_LOG_DIR               "%s/pol_%i/log"
#define DM_NEW_DIR               "%s/pol_%i/sol_%i/new"
#define DM_RAW_DIR               "%s/pol_%i/sol_%i/raw"
#define DM_DECOD_DIR             "%s/pol_%i/sol_%i/decode"
#define DM_FAULT_DIR             "%s/pol_%i/sol_%i/fault"
#define DM_HISTORY_DIR           "%s/pol_%i/sol_%i/history"
#define DM_DEMA_RUN              "/var/run/dema.pid"
#define DM_RT_START_FILE         "realtime_start"
#define DM_RT_STOP_FILE          "realtime_stop"
#define DM_DELETE_CASE           "delete"
#define DM_DELETE_SESSION        "sol_rm"
#define DM_LUCENE_CMD            "/opt/xplico/script/session_mng.pyc -l %i %i"
#define POL_INIT_SESSION_FILE    "pol_sinit.cfg"
#define POL_END_SESSION_FILE     "pol_send.cfg"
#define DB_T_MYSQL               "mysql"
#define DB_T_SQLITE              "sqlite"
#define DB_T_POSTGRES            "postgres"
/* cfg master files */
#define DM_XPLICO_LITE_CFG       "xplico_install_lite.cfg"
#define DM_XPLICO_MYSQL_CFG      "xplico_install_mysql.cfg"
#define DM_XPLICO_POSTGRES_CFG   "xplico_install_postgres.cfg"
#define DM_MFBC_LITE_CFG         "mfbc_install_lite.cfg"
#define DM_MFBC_POSTGRES_CFG     "mfbc_install_postgres.cfg"
#define DM_MFBC_MYSQL_CFG        "mfbc_install_mysql.cfg"
#define DM_MWMAIL_LITE_CFG       "mwmail_install_lite.cfg"
#define DM_MWMAIL_POSTGRES_CFG   "mwmail_install_postgres.cfg"
#define DM_MWMAIL_MYSQL_CFG      "mwmail_install_mysql.cfg"
#define DM_MFILE_LITE_CFG        "mfile_install_lite.cfg"
#define DM_MFILE_POSTGRES_CFG    "mfile_install_postgres.cfg"
#define DM_MFILE_MYSQL_CFG       "mfile_install_mysql.cfg"
#define DM_MPALTALK_EXP_LITE_CFG        "mpaltalk_install_lite.cfg"
#define DM_MPALTALK_EXP_POSTGRES_CFG    "mpaltalk_install_postgres.cfg"
#define DM_MPALTALK_EXP_MYSQL_CFG       "mpaltalk_install_mysql.cfg"
/* dema config file */
#define DM_DEFAULT_MYSQL_CFG     "/opt/xplico/cfg/dema_mysql.cfg"


/* manipulartor cfg line for xplico */
#define DM_XPLICO_MANIP          "MANIP=%s   MPHOST=127.0.0.1  MPPORT=%s"


/** boolean type */
typedef unsigned char bool;
#define TRUE     (0==0)
#define FALSE    (!TRUE)

/** erase session timeout */
#define DM_ERASE_SESSION           2
#define DM_END_TO                600 /* 10 min */
#define DM_GROWTH_TO              10 /* sec */


/** task pid list  */
typedef struct _task task;
struct _task {
    int tot;         /* total task in execution */
    pid_t xplico;    /* decoder */
    pid_t mfbc;      /* facebook chat aggregator */
    pid_t mwmail;    /* webmail manipulator */
    pid_t mfile;     /* http file aggregator */
    pid_t mpaltalk;  /* paltalk express aggregator */
};


/** pol->session & decoder  */
typedef struct _podec podec;
struct _podec {
    int pol_id;                  /* pol ID */
    int sol_id;                  /* sol ID */
    bool run;                    /* task running */
    task pid;                    /* task pid */
    bool end;                    /* closing task */
    int end_to;                  /* end timeout [s] */
    char name[DM_FILENAME_PATH]; /* file name */
    off_t size;                  /* size of last file */
    int filenum;                 /* file number */
    time_t growth;               /* file growth */
    bool rt;                     /* real time or not */
};

#endif /* __DEMA_H__ */
