/* session_decoding.c
 * Session decoding monitoring
 *
 * $Id: $
 *
 * Xplico System
 * By Gianluca Costa <g.costa@xplico.org>
 * Copyright 2007-2011 Gianluca Costa & Andrea de Franceschi. Web: www.xplico.org
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#include <sys/types.h>
#include <dirent.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>


#include "session_decoding.h"
#include "dbinterface.h"
#include "config_file.h"


static char filename[DM_FILENAME_PATH];

static int ListSort(const void *a, const void *b)
{
    return strcmp(*(char **)a, *(char **)b);
}


static unsigned short ManipPort(unsigned short free_port)
{
    struct sockaddr_in servAddr;
    int yes, sd;
    
    /* create socket */
    sd = socket(AF_INET, SOCK_STREAM, 0);
    if (sd < 0) {
        printf("cannot open socket\n");
        return 0;
    }
    yes = 1;
    if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR,
                   (char *) &yes, sizeof (yes)) < 0) {
        printf("setsockopt\n");
        close(sd);
        return 0;
    }
#ifdef SO_REUSEPORT
    if (setsockopt(sock, SOL_SOCKET, SO_REUSEPORT,
                   (char *) &yes, sizeof(yes)) < 0) {
        perror("SO_REUSEPORT");
        close(sd);
        return 0;
    }
#endif

    do {
        /* bind server port */
        memset (&servAddr, 0, sizeof (servAddr));
        servAddr.sin_family = AF_INET;
        servAddr.sin_addr.s_addr = htonl(INADDR_ANY);
        servAddr.sin_port = htons(free_port);
        
        if (bind(sd, (struct sockaddr *) &servAddr, sizeof(servAddr)) == 0) {
            break;
        }
        free_port++;
    } while (free_port != 0);
    
    close(sd);
    
    return free_port;
}


int SeDeFind(char *main_dir, podec *tbl, int dim)
{
    DIR *dir, *dir1;
    struct dirent *entry, *entry1;
    int i, tmp, pol_id, sol_id;
    int new, next;
    int len_pol, len_ses;
    char dir_path[DM_FILENAME_PATH];

    dir = opendir(main_dir);
    if (dir == NULL) {
        perror("");
        exit(-1);
        return -1;
    }

    /* first empty position in tbl */
    next = 0;
    while (tbl[next].pol_id != -1)
        next++;

    /* pol directory */
    new = 0;
    len_pol = strlen(DM_POL_NAME);
    len_ses = strlen(DM_SESSION_NAME);
    while((entry = readdir(dir)) != NULL && new < DM_TBL_ADD) {
        if (entry->d_name[0] == '.')
            continue;
        /* check if this directory is a pol directory and if alrady exist in tbl */
        if (strncmp(entry->d_name, DM_POL_NAME, len_pol) == 0) {
            if (sscanf(entry->d_name, DM_POL_NAME"%i", &pol_id) == 1) {
                for (i=0; i<dim; i++) {
                    if (tbl[i].pol_id == pol_id) {
                        break;
                    }
                }
                /* next free position */
                while (tbl[next].pol_id != -1)
                    next++;
                if (i == dim) {
                    /* session directory */
                    sprintf(dir_path, DM_POL_DIR, main_dir, pol_id);
                    dir1 = opendir(dir_path);
                    if (dir1 != NULL) {
                        sol_id = INT_MAX;
                        while((entry1 = readdir(dir1)) != NULL) {
                            if (entry->d_name[0] == '.')
                                continue;
                            /* check if this directory is a pol directory and if alrady exist in tbl */
                            if (strncmp(entry1->d_name, DM_SESSION_NAME, len_ses) == 0) {
                                if (sscanf(entry1->d_name, DM_SESSION_NAME"%i", &tmp) == 1) {
                                    if (tmp < sol_id)
                                        sol_id = tmp;
                                }
                            }
                        }
                        closedir(dir1);
                        if (sol_id != INT_MAX) {
                            /* new pol */
                            printf("New Case dir: %s\n", dir_path);
                            tbl[next].pol_id = pol_id;
                            tbl[next].sol_id = sol_id;
                            tbl[next].run = FALSE;
                            memset(&tbl[next].pid, 0, sizeof(task));
                            tbl[next].end = FALSE;
                            tbl[next].name[0] = '\0';
                            tbl[next].size = 0;
                            tbl[next].filenum = 0;
                            new++;
                        }
                    }
                }
            }
        }
    }
    closedir(dir);

    return new;
}


int SeDeStart(dbtype db_t, dbconf *db_c, char *main_dir, int pol, int session, task *pid, bool rt, char *interf, char *filter)
{
    int ptsk;
    char app_path[DM_FILENAME_PATH];
    char config_file[DM_FILENAME_PATH];
    char work_dir[DM_FILENAME_PATH];
    char end_file[DM_FILENAME_PATH];
    char cmd[2*DM_FILENAME_PATH];
    char sport_mfbc[32], sport_mwmail[32], sport_mfile[32],
        sport_mpaltalk[32];
    char *xpl_cfg, *mfbc_cfg, *mwmail_cfg, *mfile_cfg, *mpaltalk_cfg;
    bool mfbc_en, mwmail_en, mfile_en, mpaltalk_en;
    unsigned short port;
    struct stat app_info;

    /* remove any other command file */
    sprintf(end_file, DM_DECOD_DIR"/%s", main_dir, pol, session, POL_END_SESSION_FILE);
    remove(end_file);
    mfbc_en = mwmail_en = mfile_en = mpaltalk_en = FALSE;
    pid->tot = 0;
    /* cfg master files */
    switch (db_t) {
    case DB_SQLITE:
        xpl_cfg = DM_XPLICO_LITE_CFG;
        mfbc_cfg = DM_MFBC_LITE_CFG;
        mwmail_cfg = DM_MWMAIL_LITE_CFG;
        mfile_cfg = DM_MFILE_LITE_CFG;
        mpaltalk_cfg = DM_MPALTALK_EXP_LITE_CFG;
        break;

    case DB_MYSQL:
        xpl_cfg = DM_XPLICO_MYSQL_CFG;
        mfbc_cfg = DM_MFBC_MYSQL_CFG;
        mwmail_cfg = DM_MWMAIL_MYSQL_CFG;
        mfile_cfg = DM_MFILE_MYSQL_CFG;
        mpaltalk_cfg = DM_MPALTALK_EXP_MYSQL_CFG;
        break;

    case DB_POSTGRES:
        xpl_cfg = DM_XPLICO_POSTGRES_CFG;
        mfbc_cfg = DM_MFBC_POSTGRES_CFG;
        mwmail_cfg = DM_MWMAIL_POSTGRES_CFG;
        mfile_cfg = DM_MFILE_POSTGRES_CFG;
        mpaltalk_cfg = DM_MPALTALK_EXP_POSTGRES_CFG;
        break;
    }

    /* start manipulators */
    /*  facebook chat manipulator */
    sprintf(app_path, "%s/bin/mfbc", main_dir);
    if (stat(app_path, &app_info) == 0) {
        mfbc_en = TRUE;
        /*  port to use */
        port = ManipPort(23456); /* basic port */
        sprintf(sport_mfbc, "%i", port);
        ptsk = fork();
        if (ptsk == 0) {
            /* create config file */
            sprintf(config_file, "%s/cfg/%s", main_dir, mfbc_cfg);
            sprintf(work_dir, DM_TMP_DIR, main_dir, pol);
            sprintf(cmd, "cp -a %s %s", config_file, work_dir);
            system(cmd);
            sprintf(config_file, DM_TMP_DIR"/%s", main_dir, pol, mfbc_cfg);
            sprintf(cmd, "echo LOG_DIR_PATH="DM_LOG_DIR" >> %s", main_dir, pol, config_file);
            system(cmd);
            sprintf(config_file, DM_TMP_DIR"/%s", main_dir, pol, mfbc_cfg);
            sprintf(cmd, "echo TMP_DIR_PATH="DM_TMP_DIR"/mfbc >> %s", main_dir, pol, config_file);
            system(cmd);
            
            /* mfbc process */
            sprintf(work_dir, DM_DECOD_DIR, main_dir, pol, session);
            execlp(app_path, "mfbc", "-c", config_file, "-s", "-p", sport_mfbc, NULL);
            
            exit(-1);
        }
        else if (ptsk == -1) {
            return -1;
        }
        pid->mfbc = ptsk;
        pid->tot++;
    }
    
    /*  webmail manipulator */
    sprintf(app_path, "%s/bin/mwmail", main_dir);
    if (stat(app_path, &app_info) == 0) {
        mwmail_en = TRUE;
        /*  port to use */
        port = ManipPort(port + 1);
        sprintf(sport_mwmail, "%i", port);
        ptsk = fork();
        if (ptsk == 0) {
            /* create config file */
            sprintf(config_file, "%s/cfg/%s", main_dir, mwmail_cfg);
            sprintf(work_dir, DM_TMP_DIR, main_dir, pol);
            sprintf(cmd, "cp -a %s %s", config_file, work_dir);
            system(cmd);
            sprintf(config_file, DM_TMP_DIR"/%s", main_dir, pol, mwmail_cfg);
            sprintf(cmd, "echo LOG_DIR_PATH="DM_LOG_DIR" >> %s", main_dir, pol, config_file);
            system(cmd);
            sprintf(config_file, DM_TMP_DIR"/%s", main_dir, pol, mwmail_cfg);
            sprintf(cmd, "echo TMP_DIR_PATH="DM_TMP_DIR"/mwmail >> %s", main_dir, pol, config_file);
            system(cmd);
            
            /* mwmail process */
            sprintf(work_dir, DM_DECOD_DIR, main_dir, pol, session);
            execlp(app_path, "mwmail", "-c", config_file, "-s", "-p", sport_mwmail, NULL);
            
            exit(-1);
        }
        else if (ptsk == -1) {
            return -1;
        }
        pid->mwmail = ptsk;
        pid->tot++;
    }

    /*  http file (mfile) manipulator */
    sprintf(app_path, "%s/bin/mfile", main_dir);
    if (stat(app_path, &app_info) == 0) {
        mfile_en = TRUE;
        /*  port to use */
        port = ManipPort(port + 1);
        sprintf(sport_mfile, "%i", port);
        ptsk = fork();
        if (ptsk == 0) {
            /* create config file */
            sprintf(config_file, "%s/cfg/%s", main_dir, mfile_cfg);
            sprintf(work_dir, DM_TMP_DIR, main_dir, pol);
            sprintf(cmd, "cp -a %s %s", config_file, work_dir);
            system(cmd);
            sprintf(config_file, DM_TMP_DIR"/%s", main_dir, pol, mfile_cfg);
            sprintf(cmd, "echo LOG_DIR_PATH="DM_LOG_DIR" >> %s", main_dir, pol, config_file);
            system(cmd);
            sprintf(config_file, DM_TMP_DIR"/%s", main_dir, pol, mfile_cfg);
            sprintf(cmd, "echo TMP_DIR_PATH="DM_TMP_DIR"/mfile >> %s", main_dir, pol, config_file);
            system(cmd);
            
            /* mfile process */
            sprintf(work_dir, DM_DECOD_DIR, main_dir, pol, session);
            execlp(app_path, "mfile", "-c", config_file, "-s", "-p", sport_mfile, NULL);
            
            exit(-1);
        }
        else if (ptsk == -1) {
            return -1;
        }
        pid->mfile = ptsk;
        pid->tot++;
    }

    /*  paltalk express aggregator */
    sprintf(app_path, "%s/bin/mpaltalk", main_dir);
    if (stat(app_path, &app_info) == 0) {
        mpaltalk_en = TRUE;
        /*  port to use */
        port = ManipPort(port + 1);
        sprintf(sport_mpaltalk, "%i", port);
        ptsk = fork();
        if (ptsk == 0) {
            /* create config file */
            sprintf(config_file, "%s/cfg/%s", main_dir, mpaltalk_cfg);
            sprintf(work_dir, DM_TMP_DIR, main_dir, pol);
            sprintf(cmd, "cp -a %s %s", config_file, work_dir);
            system(cmd);
            sprintf(config_file, DM_TMP_DIR"/%s", main_dir, pol, mpaltalk_cfg);
            sprintf(cmd, "echo LOG_DIR_PATH="DM_LOG_DIR" >> %s", main_dir, pol, config_file);
            system(cmd);
            sprintf(config_file, DM_TMP_DIR"/%s", main_dir, pol, mpaltalk_cfg);
            sprintf(cmd, "echo TMP_DIR_PATH="DM_TMP_DIR"/mpaltalk >> %s", main_dir, pol, config_file);
            system(cmd);
            
            /* mpaltalk process */
            sprintf(work_dir, DM_DECOD_DIR, main_dir, pol, session);
            execlp(app_path, "mpaltalk", "-c", config_file, "-s", "-p", sport_mpaltalk, NULL);
            
            exit(-1);
        }
        else if (ptsk == -1) {
            return -1;
        }
        pid->mpaltalk = ptsk;
        pid->tot++;
    }

    /* start xplico */
    ptsk = fork();
    if (ptsk == 0) {
        sleep(3); /* wait manipulators... to improve */
        /* create config file */
        sprintf(config_file, "%s/cfg/%s", main_dir, xpl_cfg);
        sprintf(work_dir, DM_TMP_DIR, main_dir, pol);
        sprintf(cmd, "cp -a %s %s", config_file, work_dir);
        system(cmd);
        /* manipulators */
        /*  face book */
        if (mfbc_en) {
            sprintf(config_file, DM_TMP_DIR"/%s", main_dir, pol, xpl_cfg);
            sprintf(cmd, "echo "DM_XPLICO_MANIP" >> %s", "fbwchat", sport_mfbc, config_file);
            system(cmd);
        }
        /*  web mail */
        if (mwmail_en) {
            sprintf(config_file, DM_TMP_DIR"/%s", main_dir, pol, xpl_cfg);
            sprintf(cmd, "echo "DM_XPLICO_MANIP" >> %s", "webmail", sport_mwmail, config_file);
            system(cmd);
        }
        /*   http file */
        if (mfile_en) {
            sprintf(config_file, DM_TMP_DIR"/%s", main_dir, pol, xpl_cfg);
            sprintf(cmd, "echo "DM_XPLICO_MANIP" >> %s", "httpfd", sport_mfile, config_file);
            system(cmd);
        }
        /*   paltalk express */
        if (mpaltalk_en) {
            sprintf(config_file, DM_TMP_DIR"/%s", main_dir, pol, xpl_cfg);
            sprintf(cmd, "echo "DM_XPLICO_MANIP" >> %s", "paltalk_exp", sport_mpaltalk, config_file);
            system(cmd);
        }

        /* log, pol, sol */
        sprintf(config_file, DM_TMP_DIR"/%s", main_dir, pol, xpl_cfg);
        sprintf(cmd, "echo LOG_DIR_PATH="DM_LOG_DIR" >> %s", main_dir, pol, config_file);
        system(cmd);
        sprintf(config_file, DM_TMP_DIR"/%s", main_dir, pol, xpl_cfg);
        sprintf(cmd, "echo TMP_DIR_PATH="DM_TMP_DIR"/xplico >> %s", main_dir, pol, config_file);
        system(cmd);
        sprintf(cmd, "echo POL_ID=%i > "DM_DECOD_DIR"/%s", pol,
                main_dir, pol, session, POL_INIT_SESSION_FILE);
        system(cmd);
        sprintf(cmd, "echo SESSION_ID=%i >> "DM_DECOD_DIR"/%s", session,
                main_dir, pol, session, POL_INIT_SESSION_FILE);
        system(cmd);

        /* xplico process */
        sprintf(app_path, "%s/bin/xplico", main_dir);
        sprintf(work_dir, DM_DECOD_DIR, main_dir, pol, session);
        if (rt == FALSE) {
            execlp(app_path, "xplico", "-c", config_file, "-m", "pol", "-d", work_dir, NULL);
        }
        else {
            if (filter != NULL && filter[0] != '\0') {
                execlp(app_path, "xplico", "-c", config_file, "-m", "rltm_pol", "-r", "-i", interf, "-f", filter, "-d", work_dir, NULL);
            }
            else {
                execlp(app_path, "xplico", "-c", config_file, "-m", "rltm_pol", "-r", "-i", interf, "-d", work_dir, NULL);
            }
        }
        exit(-1);
    }
    else if (ptsk == -1) {
        return -1;
    }
    pid->xplico = ptsk;
    pid->tot++;

    return 0;
}


int SeDeEnd(char *main_dir, int pol, int session, task *pid)
{
    char end_file[DM_FILENAME_PATH];
    int fd;

    /* end xplico */
    sprintf(end_file, DM_DECOD_DIR"/%s", main_dir, pol, session, POL_END_SESSION_FILE);
    fd = open(end_file, O_CREAT|O_RDWR, 0x01B6); /* only create file */
    if (fd != -1)
        close(fd);

    return 0;
}


static char* SeDeFileSrc(char *main_dir, int pol, int session, short type)
{
    char newdir[DM_FILENAME_PATH];
    DIR *dir;
    struct dirent *entry;
    int i, num;
    char **list;
    
    switch (type) {
    case 0:
        sprintf(newdir, DM_NEW_DIR, main_dir, pol, session);
        break;

    case 1:
        sprintf(newdir, DM_DECOD_DIR, main_dir, pol, session);
        break;
    }
    
    dir = opendir(newdir);
    if (dir == NULL) {
        return NULL;
    }

    /* file list */
    num = 0;
    list = NULL;
    while((entry = readdir(dir)) != NULL) {
        if (entry->d_name[0] == '.')
            continue;
        list = realloc(list, sizeof(char *)*(num+1));
        list[num] = malloc(strlen(entry->d_name)+5);
        strcpy(list[num], entry->d_name);
        num++;
        
    }
    closedir(dir);
   
    /* sort */
    qsort(list, num, sizeof(char *), ListSort);
    if (num == 0) {
        return NULL;
    }
    if (list != NULL) {
        strcpy(filename, list[0]);
        for (i=0; i<num; i++) {
            free(list[i]);
        }
        free(list);
    }

    return filename;
}


char *SeDeFileNew(char *main_dir, int pol, int session)
{
    return SeDeFileSrc(main_dir, pol, session, 0);
}


char *SeDeFileDecode(char *main_dir, int pol, int session)
{
    return SeDeFileSrc(main_dir, pol, session, 1);
}


int SeDeNextSession(char *main_dir, int pol, int session)
{
    DIR *dir;
    struct dirent *entry;
    char dir_path[DM_FILENAME_PATH];
    int len_pol, len_ses;
    int tmp;
    int sol_id;

    sol_id = -1;

    /* session directory */
    len_pol = strlen(DM_POL_NAME);
    len_ses = strlen(DM_SESSION_NAME);
    sprintf(dir_path, DM_POL_DIR, main_dir, pol);
    dir = opendir(dir_path);
    if (dir != NULL) {
        sol_id = INT_MAX;
        while((entry = readdir(dir)) != NULL) {
            if (entry->d_name[0] == '.')
                continue;
            /* check if this directory is a pol directory and if alrady exist in tbl */
            if (strncmp(entry->d_name, DM_SESSION_NAME, len_ses) == 0) {
                if (sscanf(entry->d_name, DM_SESSION_NAME"%i", &tmp) == 1) {
                    if (tmp < sol_id && tmp > session)
                        sol_id = tmp;
                }
            }
        }
        closedir(dir);
        
        /* check value */
        if (sol_id == INT_MAX) {
            sol_id = -1;
        }
    }
    
    return sol_id;
}


int SeDeRun(task *pid, pid_t chld, bool tclear)
{
    int ret;
    
    ret = -1;
    /* check all application */
    if (pid->xplico == chld) {
        if (tclear)
            pid->xplico = 0;
        ret = 0;
    }
    else if (pid->mfbc == chld) {
        if (tclear)
            pid->mfbc = 0;
        ret = 0;
    }
    else if (pid->mwmail == chld) {
        if (tclear)
            pid->mwmail = 0;
        ret = 0;
    }
    else if (pid->mfile == chld) {
        if (tclear)
            pid->mfile = 0;
        ret = 0;
    }
    else if (pid->mpaltalk == chld) {
        if (tclear)
            pid->mpaltalk = 0;
        ret = 0;
    }

    return ret;
}


int SeDeKill(podec *tbl, int id)
{
    if (tbl[id].run == TRUE) {
        /* decoder */
        if (tbl[id].pid.xplico != 0)
            kill(tbl[id].pid.xplico, SIGKILL);
        /* manipulators */
        if (tbl[id].pid.mfbc != 0)
            kill(tbl[id].pid.mfbc, SIGKILL);
        if (tbl[id].pid.mwmail != 0)
            kill(tbl[id].pid.mwmail, SIGKILL);
        if (tbl[id].pid.mfile != 0)
            kill(tbl[id].pid.mfile, SIGKILL);
        if (tbl[id].pid.mpaltalk != 0)
            kill(tbl[id].pid.mpaltalk, SIGKILL);
    }
    
    return 0;
}

